from .widgets2d import (
    VispyPlotCanvas,
    VispyImageCanvas,
    VispyMultiPlotCanvas,
    VispyMultiImageCanvas,
)
from .widgets3d import Vispy3DCanvas

__all__ = [
    "VispyPlotCanvas",
    "VispyImageCanvas",
    "VispyMultiPlotCanvas",
    "VispyMultiImageCanvas",
    "Vispy3DCanvas",
]
